@@ -1,9 +1,9 @@
-/*  Lzip - Data compressor based on the LZMA algorithm
-    Copyright (C) 2008, 2009, 2010, 2011, 2012, 2013 Antonio Diaz Diaz.
+/*  Lzip - LZMA lossless data compressor
+    Copyright (C) 2008-2015 Antonio Diaz Diaz.
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
+    the Free Software Foundation, either version 2 of the License, or
     (at your option) any later version.
 
     This program is distributed in the hope that it will be useful,
@@ -76,6 +76,7 @@ public:
     code = 0;
     for( int i = 0; i < 5; ++i ) code = (code << 8) | get_byte();
     range = 0xFFFFFFFFU;
+    code &= range;		// make sure that first byte is discarded
     }
 
   void normalize()
@@ -120,24 +121,33 @@ public:
       }
     }
 
-  int decode_tree( Bit_model bm[], const int num_bits )
+  int decode_tree3( Bit_model bm[] )
     {
-    int model = 1;
-    for( int i = num_bits; i > 0; --i )
-      model = ( model << 1 ) | decode_bit( bm[model] );
-    return model - (1 << num_bits);
+    int symbol = 1;
+    symbol = ( symbol << 1 ) | decode_bit( bm[symbol] );
+    symbol = ( symbol << 1 ) | decode_bit( bm[symbol] );
+    symbol = ( symbol << 1 ) | decode_bit( bm[symbol] );
+    return symbol & 7;
     }
 
   int decode_tree6( Bit_model bm[] )
     {
-    int model = 1;
-    model = ( model << 1 ) | decode_bit( bm[model] );
-    model = ( model << 1 ) | decode_bit( bm[model] );
-    model = ( model << 1 ) | decode_bit( bm[model] );
-    model = ( model << 1 ) | decode_bit( bm[model] );
-    model = ( model << 1 ) | decode_bit( bm[model] );
-    model = ( model << 1 ) | decode_bit( bm[model] );
-    return model - (1 << 6);
+    int symbol = 1;
+    symbol = ( symbol << 1 ) | decode_bit( bm[symbol] );
+    symbol = ( symbol << 1 ) | decode_bit( bm[symbol] );
+    symbol = ( symbol << 1 ) | decode_bit( bm[symbol] );
+    symbol = ( symbol << 1 ) | decode_bit( bm[symbol] );
+    symbol = ( symbol << 1 ) | decode_bit( bm[symbol] );
+    symbol = ( symbol << 1 ) | decode_bit( bm[symbol] );
+    return symbol & 0x3F;
+    }
+
+  int decode_tree8( Bit_model bm[] )
+    {
+    int symbol = 1;
+    while( symbol < 0x100 )
+      symbol = ( symbol << 1 ) | decode_bit( bm[symbol] );
+    return symbol & 0xFF;
     }
 
   int decode_tree_reversed( Bit_model bm[], const int num_bits )
@@ -156,10 +166,9 @@ public:
   int decode_tree_reversed4( Bit_model bm[] )
     {
     int model = 1;
-    int symbol = 0;
+    int symbol = decode_bit( bm[model] );
+    model = (model << 1) + symbol;
     int bit = decode_bit( bm[model] );
-    model = (model << 1) + bit; symbol |= bit;
-    bit = decode_bit( bm[model] );
     model = (model << 1) + bit; symbol |= (bit << 1);
     bit = decode_bit( bm[model] );
     model = (model << 1) + bit; symbol |= (bit << 2);
@@ -171,42 +180,29 @@ public:
     {
     Bit_model * const bm1 = bm + 0x100;
     int symbol = 1;
-    for( int i = 7; i >= 0; --i )
+    while( symbol < 0x100 )
       {
       match_byte <<= 1;
       const int match_bit = match_byte & 0x100;
       const int bit = decode_bit( bm1[match_bit+symbol] );
-      symbol = ( symbol << 1 ) + bit;
+      symbol = ( symbol << 1 ) | bit;
       if( match_bit != bit << 8 )
         {
         while( symbol < 0x100 )
-          symbol = ( symbol << 1 ) + decode_bit( bm[symbol] );
+          symbol = ( symbol << 1 ) | decode_bit( bm[symbol] );
         break;
         }
       }
-    return symbol - 0x100;
+    return symbol & 0xFF;
     }
-  };
 
-
-class Len_decoder
-  {
-  Bit_model choice1;
-  Bit_model choice2;
-  Bit_model bm_low[pos_states][len_low_symbols];
-  Bit_model bm_mid[pos_states][len_mid_symbols];
-  Bit_model bm_high[len_high_symbols];
-
-public:
-  int decode( Range_decoder & range_decoder, const int pos_state )
+  int decode_len( Len_model & lm, const int pos_state )
     {
-    if( range_decoder.decode_bit( choice1 ) == 0 )
-      return range_decoder.decode_tree( bm_low[pos_state], len_low_bits );
-    if( range_decoder.decode_bit( choice2 ) == 0 )
-      return len_low_symbols +
-             range_decoder.decode_tree( bm_mid[pos_state], len_mid_bits );
-    return len_low_symbols + len_mid_symbols +
-           range_decoder.decode_tree( bm_high, len_high_bits );
+    if( decode_bit( lm.choice1 ) == 0 )
+      return decode_tree3( lm.bm_low[pos_state] );
+    if( decode_bit( lm.choice2 ) == 0 )
+      return len_low_symbols + decode_tree3( lm.bm_mid[pos_state] );
+    return len_low_symbols + len_mid_symbols + decode_tree8( lm.bm_high );
     }
   };
 
@@ -214,8 +210,8 @@ public:
 class LZ_decoder
   {
   unsigned long long partial_data_pos;
-  Range_decoder & range_decoder;
-  const int dictionary_size;
+  Range_decoder & rdec;
+  const unsigned dictionary_size;
   const int buffer_size;
   uint8_t * const buffer;	// output buffer
   int pos;			// current pos in buffer
@@ -227,13 +223,13 @@ class LZ_decoder
   void flush_data();
   bool verify_trailer( const Pretty_print & pp ) const;
 
-  uint8_t get_prev_byte() const
+  uint8_t peek_prev() const
     {
     const int i = ( ( pos > 0 ) ? pos : buffer_size ) - 1;
     return buffer[i];
     }
 
-  uint8_t get_byte( const int distance ) const
+  uint8_t peek( const int distance ) const
     {
     int i = pos - distance - 1;
     if( i < 0 ) i += buffer_size;
@@ -267,24 +263,23 @@ class LZ_decoder
   void operator=( const LZ_decoder & );		// declared as private
 
 public:
-  LZ_decoder( const File_header & header, Range_decoder & rdec, const int ofd )
+  LZ_decoder( const File_header & header, Range_decoder & rde, const int ofd )
     :
     partial_data_pos( 0 ),
-    range_decoder( rdec ),
+    rdec( rde ),
     dictionary_size( header.dictionary_size() ),
-    buffer_size( std::max( 65536, dictionary_size ) ),
+    buffer_size( std::max( 65536U, dictionary_size ) ),
     buffer( new uint8_t[buffer_size] ),
     pos( 0 ),
     stream_pos( 0 ),
     crc_( 0xFFFFFFFFU ),
     outfd( ofd ),
     member_version( header.version() )
-    { buffer[buffer_size-1] = 0; }	// prev_byte of first_byte
+    { buffer[buffer_size-1] = 0; }		// prev_byte of first byte
 
   ~LZ_decoder() { delete[] buffer; }
 
   unsigned crc() const { return crc_ ^ 0xFFFFFFFFU; }
-
   unsigned long long data_position() const { return partial_data_pos + pos; }
 
   int decode_member( const Pretty_print & pp );
