@@ -1,9 +1,9 @@
-/*  Lzip - Data compressor based on the LZMA algorithm
-    Copyright (C) 2008, 2009, 2010, 2011, 2012, 2013 Antonio Diaz Diaz.
+/*  Lzip - LZMA lossless data compressor
+    Copyright (C) 2008-2015 Antonio Diaz Diaz.
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
+    the Free Software Foundation, either version 2 of the License, or
     (at your option) any later version.
 
     This program is distributed in the hope that it will be useful,
@@ -15,7 +15,7 @@
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 /*
-    Return values: 0 for a normal exit, 1 for environmental problems
+    Exit status: 0 for a normal exit, 1 for environmental problems
     (file not found, invalid flags, I/O errors, etc), 2 to indicate a
     corrupt or invalid input file, 3 for an internal consistency error
     (eg, bug) which caused lzip to panic.
@@ -56,9 +56,14 @@
 #include "arg_parser.h"
 #include "lzip.h"
 #include "decoder.h"
+#include "encoder_base.h"
 #include "encoder.h"
 #include "fast_encoder.h"
 
+#ifndef O_BINARY
+#define O_BINARY 0
+#endif
+
 #if CHAR_BIT != 8
 #error "Environments where CHAR_BIT != 8 are not supported."
 #endif
@@ -68,15 +73,9 @@ namespace {
 
 const char * const Program_name = "Lzip";
 const char * const program_name = "lzip";
-const char * const program_year = "2013";
+const char * const program_year = "2015";
 const char * invocation_name = 0;
 
-#ifdef O_BINARY
-const int o_binary = O_BINARY;
-#else
-const int o_binary = 0;
-#endif
-
 struct { const char * from; const char * to; } const known_extensions[] = {
   { ".lz",  ""     },
   { ".tlz", ".tar" },
@@ -84,17 +83,14 @@ struct { const char * from; const char * to; } const known_extensions[] = {
 
 struct Lzma_options
   {
-  int dictionary_size;		// 4KiB..512MiB
-  int match_len_limit;		// 5..273
+  int dictionary_size;		// 4 KiB .. 512 MiB
+  int match_len_limit;		// 5 .. 273
   };
 
 enum Mode { m_compress, m_decompress, m_test };
-const unsigned long long max_member_size = 0x1000000000000000ULL;
-const unsigned long long max_volume_size = 0x7FFFFFFFFFFFFFFFULL;
 
 std::string output_filename;
 int outfd = -1;
-int verbosity = 0;
 const mode_t usr_rw = S_IRUSR | S_IWUSR;
 const mode_t all_rw = usr_rw | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH;
 mode_t outfd_mode = usr_rw;
@@ -103,7 +99,7 @@ bool delete_output_on_interrupt = false;
 
 void show_help()
   {
-  std::printf( "%s - Data compressor based on the LZMA algorithm.\n", Program_name );
+  std::printf( "%s - LZMA lossless data compressor.\n", Program_name );
   std::printf( "\nUsage: %s [options] [files]\n", invocation_name );
   std::printf( "\nOptions:\n"
                "  -h, --help                     display this help and exit\n"
@@ -112,12 +108,12 @@ void show_help()
                "  -c, --stdout                   send output to standard output\n"
                "  -d, --decompress               decompress\n"
                "  -f, --force                    overwrite existing output files\n"
-               "  -F, --recompress               force recompression of compressed files\n"
+               "  -F, --recompress               force re-compression of compressed files\n"
                "  -k, --keep                     keep (don't delete) input files\n"
                "  -m, --match-length=<bytes>     set match length limit in bytes [36]\n"
                "  -o, --output=<file>            if reading stdin, place the output into <file>\n"
                "  -q, --quiet                    suppress all messages\n"
-               "  -s, --dictionary-size=<bytes>  set dictionary size limit in bytes [8MiB]\n"
+               "  -s, --dictionary-size=<bytes>  set dictionary size limit in bytes [8 MiB]\n"
                "  -S, --volume-size=<bytes>      set volume size limit in bytes\n"
                "  -t, --test                     test compressed file integrity\n"
                "  -v, --verbose                  be verbose (a 2nd -v gives more)\n"
@@ -132,6 +128,10 @@ void show_help()
                "scale optimal for all files. If your files are large, very repetitive,\n"
                "etc, you may need to use the --match-length and --dictionary-size\n"
                "options directly to achieve optimal performance.\n"
+               "\nExit status: 0 for a normal exit, 1 for environmental problems (file\n"
+               "not found, invalid flags, I/O errors, etc), 2 to indicate a corrupt or\n"
+               "invalid input file, 3 for an internal consistency error (eg, bug) which\n"
+               "caused lzip to panic.\n"
                "\nReport bugs to lzip-bug@nongnu.org\n"
                "Lzip home page: http://www.nongnu.org/lzip/lzip.html\n" );
   }
@@ -139,9 +139,9 @@ void show_help()
 
 void show_version()
   {
-  std::printf( "%s %s\n", Program_name, PROGVERSION );
+  std::printf( "%s %s\n", program_name, PROGVERSION );
   std::printf( "Copyright (C) %s Antonio Diaz Diaz.\n", program_year );
-  std::printf( "License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>\n"
+  std::printf( "License GPLv2+: GNU GPL version 2 or later <http://gnu.org/licenses/gpl.html>\n"
                "This is free software: you are free to change and redistribute it.\n"
                "There is NO WARRANTY, to the extent permitted by law.\n" );
   }
@@ -149,19 +149,23 @@ void show_version()
 
 void show_header( const File_header & header )
   {
-  const char * const prefix[8] =
-    { "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi", "Yi" };
-  enum { factor = 1024 };
-  const char * p = "";
-  const char * np = "  ";
-  unsigned num = header.dictionary_size();
-  bool exact = ( num % factor == 0 );
+  if( verbosity >= 3 )
+    {
+    const char * const prefix[8] =
+      { "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi", "Yi" };
+    enum { factor = 1024 };
+    const char * p = "";
+    const char * np = "  ";
+    unsigned num = header.dictionary_size();
+    bool exact = ( num % factor == 0 );
 
-  for( int i = 0; i < 8 && ( num > 9999 || ( exact && num >= factor ) ); ++i )
-    { num /= factor; if( num % factor != 0 ) exact = false;
-      p = prefix[i]; np = ""; }
-  std::fprintf( stderr, "version %d, dictionary size %s%4u %sB.  ",
-                header.version(), np, num, p );
+    for( int i = 0; i < 8 && ( num > 9999 || ( exact && num >= factor ) ); ++i )
+      { num /= factor; if( num % factor != 0 ) exact = false;
+        p = prefix[i]; np = ""; }
+    if( verbosity >= 4 && header.version() != 1 )
+      std::fprintf( stderr, "version %d, ", header.version() );
+    std::fprintf( stderr, "dictionary size %s%4u %sB.  ", np, num, p );
+    }
   }
 
 
@@ -169,8 +173,8 @@ unsigned long long getnum( const char * const ptr,
                            const unsigned long long llimit,
                            const unsigned long long ulimit )
   {
-  errno = 0;
   char * tail;
+  errno = 0;
   unsigned long long result = strtoull( ptr, &tail, 0 );
   if( tail == ptr )
     {
@@ -180,7 +184,7 @@ unsigned long long getnum( const char * const ptr,
 
   if( !errno && tail[0] )
     {
-    int factor = ( tail[1] == 'i' ) ? 1024 : 1000;
+    const int factor = ( tail[1] == 'i' ) ? 1024 : 1000;
     int exponent = 0;
     bool bad_multiplier = false;
     switch( tail[0] )
@@ -223,7 +227,7 @@ unsigned long long getnum( const char * const ptr,
 int get_dict_size( const char * const arg )
   {
   char * tail;
-  int bits = std::strtol( arg, &tail, 0 );
+  const int bits = std::strtol( arg, &tail, 0 );
   if( bits >= min_dictionary_bits &&
       bits <= max_dictionary_bits && *tail == 0 )
     return ( 1 << bits );
@@ -244,7 +248,7 @@ int extension_index( const std::string & name )
   }
 
 
-int open_instream( const std::string & name, struct stat * const in_statsp,
+int open_instream( const char * const name, struct stat * const in_statsp,
                    const Mode program_mode, const int eindex,
                    const bool recompress, const bool to_stdout )
   {
@@ -253,17 +257,16 @@ int open_instream( const std::string & name, struct stat * const in_statsp,
     {
     if( verbosity >= 0 )
       std::fprintf( stderr, "%s: Input file '%s' already has '%s' suffix.\n",
-                    program_name, name.c_str(),
-                    known_extensions[eindex].from );
+                    program_name, name, known_extensions[eindex].from );
     }
   else
     {
-    infd = open( name.c_str(), O_RDONLY | o_binary );
+    infd = open( name, O_RDONLY | O_BINARY );
     if( infd < 0 )
       {
       if( verbosity >= 0 )
-        std::fprintf( stderr, "%s: Can't open input file '%s': %s.\n",
-                      program_name, name.c_str(), std::strerror( errno ) );
+        std::fprintf( stderr, "%s: Can't open input file '%s': %s\n",
+                      program_name, name, std::strerror( errno ) );
       }
     else
       {
@@ -272,13 +275,14 @@ int open_instream( const std::string & name, struct stat * const in_statsp,
       const bool can_read = ( i == 0 &&
                               ( S_ISBLK( mode ) || S_ISCHR( mode ) ||
                                 S_ISFIFO( mode ) || S_ISSOCK( mode ) ) );
-      if( i != 0 || ( !S_ISREG( mode ) && ( !to_stdout || !can_read ) ) )
+      const bool no_ofile = to_stdout || program_mode == m_test;
+      if( i != 0 || ( !S_ISREG( mode ) && ( !can_read || !no_ofile ) ) )
         {
         if( verbosity >= 0 )
           std::fprintf( stderr, "%s: Input file '%s' is not a regular file%s.\n",
-                        program_name, name.c_str(),
-                        ( can_read && !to_stdout ) ?
-                        " and '--stdout' was not specified" : "" );
+                        program_name, name,
+                        ( can_read && !no_ofile ) ?
+                        ",\n      and '--stdout' was not specified" : "" );
         close( infd );
         infd = -1;
         }
@@ -310,14 +314,14 @@ void set_d_outname( const std::string & name, const int i )
     }
   output_filename = name; output_filename += ".out";
   if( verbosity >= 1 )
-    std::fprintf( stderr, "%s: Can't guess original name for '%s' -- using '%s'.\n",
+    std::fprintf( stderr, "%s: Can't guess original name for '%s' -- using '%s'\n",
                   program_name, name.c_str(), output_filename.c_str() );
   }
 
 
 bool open_outstream( const bool force )
   {
-  int flags = O_CREAT | O_WRONLY | o_binary;
+  int flags = O_CREAT | O_WRONLY | O_BINARY;
   if( force ) flags |= O_TRUNC; else flags |= O_EXCL;
 
   outfd = open( output_filename.c_str(), flags, outfd_mode );
@@ -327,7 +331,7 @@ bool open_outstream( const bool force )
       std::fprintf( stderr, "%s: Output file '%s' already exists, skipping.\n",
                     program_name, output_filename.c_str() );
     else
-      std::fprintf( stderr, "%s: Can't create output file '%s': %s.\n",
+      std::fprintf( stderr, "%s: Can't create output file '%s': %s\n",
                     program_name, output_filename.c_str(), std::strerror( errno ) );
     }
   return ( outfd >= 0 );
@@ -336,7 +340,7 @@ bool open_outstream( const bool force )
 
 bool check_tty( const int infd, const Mode program_mode )
   {
-  if( program_mode == m_compress && outfd >= 0 && isatty( outfd ) )
+  if( program_mode == m_compress && isatty( outfd ) )
     {
     show_error( "I won't write compressed data to a terminal.", 0, true );
     return false;
@@ -373,10 +377,14 @@ void close_and_set_permissions( const struct stat * const in_statsp )
   bool warning = false;
   if( in_statsp )
     {
+    const mode_t mode = in_statsp->st_mode;
     // fchown will in many cases return with EPERM, which can be safely ignored.
-    if( ( fchown( outfd, in_statsp->st_uid, in_statsp->st_gid ) != 0 &&
-          errno != EPERM ) ||
-        fchmod( outfd, in_statsp->st_mode ) != 0 ) warning = true;
+    if( fchown( outfd, in_statsp->st_uid, in_statsp->st_gid ) == 0 )
+      { if( fchmod( outfd, mode ) != 0 ) warning = true; }
+    else
+      if( errno != EPERM ||
+          fchmod( outfd, mode & ~( S_ISUID | S_ISGID | S_ISVTX ) ) != 0 )
+        warning = true;
     }
   if( close( outfd ) != 0 ) cleanup_and_fail( 1 );
   outfd = -1;
@@ -395,9 +403,9 @@ void close_and_set_permissions( const struct stat * const in_statsp )
 
 bool next_filename()
   {
-  const unsigned len = std::strlen( known_extensions[0].from );
-  if( output_filename.size() >= len + 5 )		// "*00001.lz"
-    for( int i = output_filename.size() - len - 1, j = 0; j < 5; --i, ++j )
+  const unsigned ext_len = std::strlen( known_extensions[0].from );
+  if( output_filename.size() >= ext_len + 5 )		// "*00001.lz"
+    for( int i = output_filename.size() - ext_len - 1, j = 0; j < 5; --i, ++j )
       {
       if( output_filename[i] < '9' ) { ++output_filename[i]; return true; }
       else output_filename[i] = '0';
@@ -407,39 +415,44 @@ bool next_filename()
 
 
 int compress( const unsigned long long member_size,
-              const unsigned long long volume_size,
-              const Lzma_options & encoder_options, const int infd,
-              const Pretty_print & pp, const struct stat * const in_statsp )
+              const unsigned long long volume_size, const int infd,
+              const Lzma_options & encoder_options, const Pretty_print & pp,
+              const struct stat * const in_statsp, const bool zero )
   {
+  const unsigned long long cfile_size =
+    (in_statsp && S_ISREG( in_statsp->st_mode )) ? in_statsp->st_size / 100 : 0;
   int retval = 0;
-  File_header header;
-  header.set_magic();
-
+  LZ_encoder_base * encoder = 0;		// polymorphic encoder
   if( verbosity >= 1 ) pp();
-  if( !header.dictionary_size( encoder_options.dictionary_size ) ||
-      encoder_options.match_len_limit < min_match_len_limit ||
-      encoder_options.match_len_limit > max_match_len )
-    internal_error( "invalid argument to encoder" );
 
   try {
-    Matchfinder matchfinder( header.dictionary_size(),
-                             encoder_options.match_len_limit, infd );
-    header.dictionary_size( matchfinder.dictionary_size() );
+    if( zero )
+      encoder = new FLZ_encoder( infd, outfd );
+    else
+      {
+      File_header header;
+      if( header.dictionary_size( encoder_options.dictionary_size ) &&
+          encoder_options.match_len_limit >= min_match_len_limit &&
+          encoder_options.match_len_limit <= max_match_len )
+        encoder = new LZ_encoder( header.dictionary_size(),
+                                  encoder_options.match_len_limit, infd, outfd );
+      else internal_error( "invalid argument to encoder." );
+      }
 
     unsigned long long in_size = 0, out_size = 0, partial_volume_size = 0;
     while( true )		// encode one member per iteration
       {
-      LZ_encoder encoder( matchfinder, header, outfd );
-      const unsigned long long size = ( ( volume_size > 0 ) ?
-        std::min( member_size, volume_size - partial_volume_size ) : member_size );
-      if( !encoder.encode_member( size ) )
-        { pp( "Encoder error" ); retval = 1; break; }
-      in_size += matchfinder.data_position();
-      out_size += encoder.member_position();
-      if( matchfinder.finished() ) break;
+      const unsigned long long size = ( volume_size > 0 ) ?
+        std::min( member_size, volume_size - partial_volume_size ) : member_size;
+      show_progress( in_size, encoder, &pp, cfile_size );	// init
+      if( !encoder->encode_member( size ) )
+        { pp( "Encoder error." ); retval = 1; break; }
+      in_size += encoder->data_position();
+      out_size += encoder->member_position();
+      if( encoder->data_finished() ) break;
       if( volume_size > 0 )
         {
-        partial_volume_size += encoder.member_position();
+        partial_volume_size += encoder->member_position();
         if( partial_volume_size >= volume_size - min_dictionary_size )
           {
           partial_volume_size = 0;
@@ -447,19 +460,19 @@ int compress( const unsigned long long member_size,
             {
             close_and_set_permissions( in_statsp );
             if( !next_filename() )
-              { pp( "Too many volume files" ); retval = 1; break; }
+              { pp( "Too many volume files." ); retval = 1; break; }
             if( !open_outstream( true ) ) { retval = 1; break; }
             delete_output_on_interrupt = true;
             }
           }
         }
-      matchfinder.reset();
+      encoder->reset();
       }
 
     if( retval == 0 && verbosity >= 1 )
       {
       if( in_size == 0 || out_size == 0 )
-        std::fprintf( stderr, " no data compressed.\n" );
+        std::fputs( " no data compressed.\n", stderr );
       else
         std::fprintf( stderr, "%6.3f:1, %6.3f bits/byte, "
                               "%5.2f%% saved, %llu in, %llu out.\n",
@@ -471,77 +484,11 @@ int compress( const unsigned long long member_size,
     }
   catch( std::bad_alloc )
     {
-    pp( "Not enough memory. Try a smaller dictionary size" );
-    retval = 1;
-    }
-  catch( Error e ) { pp(); show_error( e.msg, errno ); retval = 1; }
-  return retval;
-  }
-
-
-int fcompress( const unsigned long long member_size,
-               const unsigned long long volume_size, const int infd,
-               const Pretty_print & pp, const struct stat * const in_statsp )
-  {
-  int retval = 0;
-  File_header header;
-  header.set_magic();
-  if( verbosity >= 1 ) pp();
-
-  try {
-    Fmatchfinder fmatchfinder( infd );
-    if( !header.dictionary_size( fmatchfinder.dictionary_size() ) )
-      internal_error( "invalid argument to encoder" );
-
-    unsigned long long in_size = 0, out_size = 0, partial_volume_size = 0;
-    while( true )		// encode one member per iteration
-      {
-      FLZ_encoder encoder( fmatchfinder, header, outfd );
-      const unsigned long long size = ( ( volume_size > 0 ) ?
-        std::min( member_size, volume_size - partial_volume_size ) : member_size );
-      if( !encoder.encode_member( size ) )
-        { pp( "Encoder error" ); retval = 1; break; }
-      in_size += fmatchfinder.data_position();
-      out_size += encoder.member_position();
-      if( fmatchfinder.finished() ) break;
-      if( volume_size > 0 )
-        {
-        partial_volume_size += encoder.member_position();
-        if( partial_volume_size >= volume_size - min_dictionary_size )
-          {
-          partial_volume_size = 0;
-          if( delete_output_on_interrupt )
-            {
-            close_and_set_permissions( in_statsp );
-            if( !next_filename() )
-              { pp( "Too many volume files" ); retval = 1; break; }
-            if( !open_outstream( true ) ) { retval = 1; break; }
-            delete_output_on_interrupt = true;
-            }
-          }
-        }
-      fmatchfinder.reset();
-      }
-
-    if( retval == 0 && verbosity >= 1 )
-      {
-      if( in_size == 0 || out_size == 0 )
-        std::fprintf( stderr, " no data compressed.\n" );
-      else
-        std::fprintf( stderr, "%6.3f:1, %6.3f bits/byte, "
-                              "%5.2f%% saved, %llu in, %llu out.\n",
-                      (double)in_size / out_size,
-                      ( 8.0 * out_size ) / in_size,
-                      100.0 * ( 1.0 - ( (double)out_size / in_size ) ),
-                      in_size, out_size );
-      }
-    }
-  catch( std::bad_alloc )
-    {
-    pp( "Not enough memory. Try a smaller dictionary size" );
+    pp( "Not enough memory. Try a smaller dictionary size." );
     retval = 1;
     }
   catch( Error e ) { pp(); show_error( e.msg, errno ); retval = 1; }
+  delete encoder;
   return retval;
   }
 
@@ -578,6 +525,7 @@ void show_trailing_garbage( const uint8_t * const data, const int size,
       garbage_msg += xdigit( data[i] & 0x0F );
       }
     }
+  garbage_msg += '.';
   pp( garbage_msg.c_str() );
   }
 
@@ -597,7 +545,7 @@ int decompress( const int infd, const Pretty_print & pp, const bool testing )
       if( rdec.finished() )			// End Of File
         {
         if( first_member )
-          { pp( "Error reading member header" ); retval = 1; }
+          { pp( "File ends unexpectedly at member header." ); retval = 2; }
         else if( verbosity >= 4 && size > 0 )
           show_trailing_garbage( header.data, size, pp, true );
         break;
@@ -605,7 +553,7 @@ int decompress( const int infd, const Pretty_print & pp, const bool testing )
       if( !header.verify_magic() )
         {
         if( first_member )
-          { pp( "Bad magic number (file not in lzip format)" ); retval = 2; }
+          { pp( "Bad magic number (file not in lzip format)." ); retval = 2; }
         else if( verbosity >= 4 )
           show_trailing_garbage( header.data, size, pp, false );
         break;
@@ -618,12 +566,13 @@ int decompress( const int infd, const Pretty_print & pp, const bool testing )
                           header.version() ); }
         retval = 2; break;
         }
-      if( header.dictionary_size() < min_dictionary_size ||
-          header.dictionary_size() > max_dictionary_size )
-        { pp( "Invalid dictionary size in member header" ); retval = 2; break; }
+      const unsigned dictionary_size = header.dictionary_size();
+      if( dictionary_size < min_dictionary_size ||
+          dictionary_size > max_dictionary_size )
+        { pp( "Invalid dictionary size in member header." ); retval = 2; break; }
 
       if( verbosity >= 2 || ( verbosity == 1 && first_member ) )
-        { pp(); if( verbosity >= 2 ) show_header( header ); }
+        { pp(); show_header( header ); }
 
       LZ_decoder decoder( header, rdec, outfd );
       const int result = decoder.decode_member( pp );
@@ -633,29 +582,20 @@ int decompress( const int infd, const Pretty_print & pp, const bool testing )
         if( verbosity >= 0 && result <= 2 )
           {
           pp();
-          if( result == 2 )
-            std::fprintf( stderr, "File ends unexpectedly at pos %llu\n",
-                          partial_file_pos );
-          else
-            std::fprintf( stderr, "Decoder error at pos %llu\n",
-                          partial_file_pos );
+          std::fprintf( stderr, "%s at pos %llu\n", ( result == 2 ) ?
+                        "File ends unexpectedly" : "Decoder error",
+                        partial_file_pos );
           }
         retval = 2; break;
         }
       if( verbosity >= 2 )
-        { if( testing ) std::fprintf( stderr, "ok\n" );
-          else std::fprintf( stderr, "done\n" ); pp.reset(); }
+        { std::fputs( testing ? "ok\n" : "done\n", stderr ); pp.reset(); }
       }
     }
-  catch( std::bad_alloc )
-    {
-    pp( "Not enough memory. Find a machine with more memory" );
-    retval = 1;
-    }
+  catch( std::bad_alloc ) { pp( "Not enough memory." ); retval = 1; }
   catch( Error e ) { pp(); show_error( e.msg, errno ); retval = 1; }
   if( verbosity == 1 && retval == 0 )
-    { if( testing ) std::fprintf( stderr, "ok\n" );
-      else std::fprintf( stderr, "done\n" ); }
+    std::fputs( testing ? "ok\n" : "done\n", stderr );
   return retval;
   }
 
@@ -677,6 +617,9 @@ void set_signals()
 } // end namespace
 
 
+int verbosity = 0;
+
+
 void show_error( const char * const msg, const int errcode, const bool help )
   {
   if( verbosity >= 0 )
@@ -686,7 +629,7 @@ void show_error( const char * const msg, const int errcode, const bool help )
       std::fprintf( stderr, "%s: %s", program_name, msg );
       if( errcode > 0 )
         std::fprintf( stderr, ": %s", std::strerror( errcode ) );
-      std::fprintf( stderr, "\n" );
+      std::fputc( '\n', stderr );
       }
     if( help )
       std::fprintf( stderr, "Try '%s --help' for more information.\n",
@@ -698,15 +641,41 @@ void show_error( const char * const msg, const int errcode, const bool help )
 void internal_error( const char * const msg )
   {
   if( verbosity >= 0 )
-    std::fprintf( stderr, "%s: internal error: %s.\n", program_name, msg );
+    std::fprintf( stderr, "%s: internal error: %s\n", program_name, msg );
   std::exit( 3 );
   }
 
 
+void show_progress( const unsigned long long partial_size,
+                    const Matchfinder_base * const m,
+                    const Pretty_print * const p,
+                    const unsigned long long cfile_size )
+  {
+  static unsigned long long csize = 0;		// file_size / 100
+  static unsigned long long psize = 0;
+  static const Matchfinder_base * mb = 0;
+  static const Pretty_print * pp = 0;
+
+  if( verbosity >= 2 )
+    {
+    if( m )					// initialize static vars
+      { csize = cfile_size; psize = partial_size; mb = m; pp = p; }
+    if( mb && pp )
+      {
+      const unsigned long long pos = psize + mb->data_position();
+      if( csize > 0 )
+        std::fprintf( stderr, "%4llu%%", pos / csize );
+      std::fprintf( stderr, "  %.1f MB\r", pos / 1000000.0 );
+      pp->reset(); (*pp)();			// restore cursor position
+      }
+    }
+  }
+
+
 int main( const int argc, const char * const argv[] )
   {
-  // Mapping from gzip/bzip2 style 1..9 compression modes
-  // to the corresponding LZMA compression modes.
+  /* Mapping from gzip/bzip2 style 1..9 compression modes
+     to the corresponding LZMA compression modes. */
   const Lzma_options option_mapping[] =
     {
     { 1 << 16,  16 },		// -0 entry values not used
@@ -720,6 +689,8 @@ int main( const int argc, const char * const argv[] )
     { 3 << 23, 132 },		// -8
     { 1 << 25, 273 } };		// -9
   Lzma_options encoder_options = option_mapping[6];	// default = "-6"
+  const unsigned long long max_member_size = 0x0008000000000000ULL;
+  const unsigned long long max_volume_size = 0x4000000000000000ULL;
   unsigned long long member_size = max_member_size;
   unsigned long long volume_size = 0;
   std::string input_filename;
@@ -754,6 +725,7 @@ int main( const int argc, const char * const argv[] )
     { 'h', "help",            Arg_parser::no  },
     { 'k', "keep",            Arg_parser::no  },
     { 'm', "match-length",    Arg_parser::yes },
+    { 'n', "threads",         Arg_parser::yes },
     { 'o', "output",          Arg_parser::yes },
     { 'q', "quiet",           Arg_parser::no  },
     { 's', "dictionary-size", Arg_parser::yes },
@@ -772,15 +744,15 @@ int main( const int argc, const char * const argv[] )
     {
     const int code = parser.code( argind );
     if( !code ) break;					// no more options
-    const char * const arg = parser.argument( argind ).c_str();
+    const std::string & arg = parser.argument( argind );
     switch( code )
       {
-      case '0': zero = true; break;
-      case '1': case '2': case '3': case '4':
+      case '0': case '1': case '2': case '3': case '4':
       case '5': case '6': case '7': case '8': case '9':
-                zero = false;
+                zero = ( code == '0' );
                 encoder_options = option_mapping[code-'0']; break;
-      case 'b': member_size = getnum( arg, 100000, max_member_size ); break;
+      case 'b': member_size = getnum( arg.c_str(), 100000, max_member_size );
+                break;
       case 'c': to_stdout = true; break;
       case 'd': program_mode = m_decompress; break;
       case 'f': force = true; break;
@@ -788,17 +760,19 @@ int main( const int argc, const char * const argv[] )
       case 'h': show_help(); return 0;
       case 'k': keep_input_files = true; break;
       case 'm': encoder_options.match_len_limit =
-                  getnum( arg, min_match_len_limit, max_match_len );
+                  getnum( arg.c_str(), min_match_len_limit, max_match_len );
                 zero = false; break;
+      case 'n': break;
       case 'o': default_output_filename = arg; break;
       case 'q': verbosity = -1; break;
-      case 's': encoder_options.dictionary_size = get_dict_size( arg );
+      case 's': encoder_options.dictionary_size = get_dict_size( arg.c_str() );
                 zero = false; break;
-      case 'S': volume_size = getnum( arg, 100000, max_volume_size ); break;
+      case 'S': volume_size = getnum( arg.c_str(), 100000, max_volume_size );
+                break;
       case 't': program_mode = m_test; break;
       case 'v': if( verbosity < 4 ) ++verbosity; break;
       case 'V': show_version(); return 0;
-      default : internal_error( "uncaught option" );
+      default : internal_error( "uncaught option." );
       }
     } // end process options
 
@@ -827,7 +801,7 @@ int main( const int argc, const char * const argv[] )
       ( filenames_given || default_output_filename.size() ) )
     set_signals();
 
-  Pretty_print pp( filenames, verbosity );
+  Pretty_print pp( filenames );
 
   int retval = 0;
   for( unsigned i = 0; i < filenames.size(); ++i )
@@ -835,13 +809,13 @@ int main( const int argc, const char * const argv[] )
     struct stat in_stats;
     output_filename.clear();
 
-    if( !filenames[i].size() || filenames[i] == "-" )
+    if( filenames[i].empty() || filenames[i] == "-" )
       {
       input_filename.clear();
       infd = STDIN_FILENO;
       if( program_mode != m_test )
         {
-        if( to_stdout || !default_output_filename.size() )
+        if( to_stdout || default_output_filename.empty() )
           outfd = STDOUT_FILENO;
         else
           {
@@ -862,7 +836,7 @@ int main( const int argc, const char * const argv[] )
       {
       input_filename = filenames[i];
       const int eindex = extension_index( input_filename );
-      infd = open_instream( input_filename, &in_stats, program_mode,
+      infd = open_instream( input_filename.c_str(), &in_stats, program_mode,
                             eindex, recompress, to_stdout );
       if( infd < 0 ) { if( retval < 1 ) retval = 1; continue; }
       if( program_mode != m_test )
@@ -892,13 +866,8 @@ int main( const int argc, const char * const argv[] )
     pp.set_name( input_filename );
     int tmp;
     if( program_mode == m_compress )
-      {
-      if( zero )
-        tmp = fcompress( member_size, volume_size, infd, pp, in_statsp );
-      else
-        tmp = compress( member_size, volume_size, encoder_options, infd,
-                        pp, in_statsp );
-      }
+      tmp = compress( member_size, volume_size, infd, encoder_options, pp,
+                      in_statsp, zero );
     else
       tmp = decompress( infd, pp, program_mode == m_test );
     if( tmp > retval ) retval = tmp;
